# coding: utf-8
"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

from unittest.mock import ANY
from unittest import IsolatedAsyncioTestCase
from mock import patch
from datetime import datetime

import urllib3

from openfga_sdk.sync.client.client import OpenFgaClient
from openfga_sdk.sync import rest
from openfga_sdk.client import ClientConfiguration
from openfga_sdk.client.models.assertion import ClientAssertion
from openfga_sdk.client.models.check_request import ClientCheckRequest
from openfga_sdk.client.models.tuple import ClientTuple
from openfga_sdk.client.models.write_request import ClientWriteRequest
from openfga_sdk.client.models.expand_request import ClientExpandRequest
from openfga_sdk.client.models.list_objects_request import ClientListObjectsRequest
from openfga_sdk.client.models.list_relations_request import ClientListRelationsRequest
from openfga_sdk.client.models.read_changes_request import ClientReadChangesRequest
from openfga_sdk.client.models.write_single_response import ClientWriteSingleResponse
from openfga_sdk.client.models.write_transaction_opts import WriteTransactionOpts
from openfga_sdk.exceptions import ValidationException, FgaValidationException, UnauthorizedException
from openfga_sdk.models.assertion import Assertion
from openfga_sdk.models.authorization_model import AuthorizationModel
from openfga_sdk.models.check_response import CheckResponse
from openfga_sdk.models.create_store_request import CreateStoreRequest
from openfga_sdk.models.create_store_response import CreateStoreResponse
from openfga_sdk.models.expand_response import ExpandResponse
from openfga_sdk.models.get_store_response import GetStoreResponse
from openfga_sdk.models.leaf import Leaf
from openfga_sdk.models.list_objects_response import ListObjectsResponse
from openfga_sdk.models.list_stores_response import ListStoresResponse
from openfga_sdk.models.node import Node
from openfga_sdk.models.object_relation import ObjectRelation
from openfga_sdk.models.read_assertions_response import ReadAssertionsResponse
from openfga_sdk.models.read_authorization_model_response import ReadAuthorizationModelResponse
from openfga_sdk.models.read_authorization_models_response import ReadAuthorizationModelsResponse
from openfga_sdk.models.read_changes_response import ReadChangesResponse
from openfga_sdk.models.read_response import ReadResponse
from openfga_sdk.models.store import Store
from openfga_sdk.models.tuple import Tuple
from openfga_sdk.models.tuple_change import TupleChange
from openfga_sdk.models.tuple_key import TupleKey
from openfga_sdk.models.tuple_operation import TupleOperation
from openfga_sdk.models.type_definition import TypeDefinition
from openfga_sdk.models.users import Users
from openfga_sdk.models.userset import Userset
from openfga_sdk.models.usersets import Usersets
from openfga_sdk.models.userset_tree import UsersetTree
from openfga_sdk.models.validation_error_message_response import ValidationErrorMessageResponse
from openfga_sdk.models.write_authorization_model_request import WriteAuthorizationModelRequest
from openfga_sdk.models.write_authorization_model_response import WriteAuthorizationModelResponse


store_id = '01YCP46JKYM8FJCQ37NMBYHE5X'
request_id = 'x1y2z3'

# Helper function to construct mock response


def http_mock_response(body, status):
    headers = urllib3.response.HTTPHeaderDict({
        'content-type': 'application/json',
        'Fga-Request-Id': request_id
    })
    return urllib3.HTTPResponse(
        body.encode('utf-8'),
        headers,
        status,
        preload_content=False
    )


def mock_response(body, status):
    obj = http_mock_response(body, status)
    return rest.RESTResponse(obj, obj.data)


class TestOpenFgaClient(IsolatedAsyncioTestCase):
    """Test for OpenFGA Client"""

    def setUp(self):
        self.configuration = ClientConfiguration(
            api_scheme='http',
            api_host="api.fga.example",
        )

    def tearDown(self):
        pass

    @patch.object(rest.RESTClientObject, 'request')
    def test_list_stores(self, mock_request):
        """Test case for list_stores

        Get all stores  # noqa: E501
        """
        response_body = '''
{
  "stores": [
    {
      "id": "01YCP46JKYM8FJCQ37NMBYHE5X",
      "name": "store1",
      "created_at": "2022-07-25T21:15:37.524Z",
      "updated_at": "2022-07-25T21:15:37.524Z",
      "deleted_at": "2022-07-25T21:15:37.524Z"
    },
    {
      "id": "01YCP46JKYM8FJCQ37NMBYHE6X",
      "name": "store2",
      "created_at": "2022-07-25T21:15:37.524Z",
      "updated_at": "2022-07-25T21:15:37.524Z",
      "deleted_at": "2022-07-25T21:15:37.524Z"
    }
  ],
  "continuation_token": "eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ=="
}
            '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.list_stores(
                options={"page_size": 1, "continuation_token": "continuation_token_example"}
            )
            self.assertIsInstance(api_response, ListStoresResponse)
            self.assertEqual(api_response.continuation_token,
                             "eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ==")
            store1 = Store(
                id="01YCP46JKYM8FJCQ37NMBYHE5X",
                name="store1",
                created_at=datetime.fromisoformat("2022-07-25T21:15:37.524+00:00"),
                updated_at=datetime.fromisoformat("2022-07-25T21:15:37.524+00:00"),
                deleted_at=datetime.fromisoformat("2022-07-25T21:15:37.524+00:00"),
            )
            store2 = Store(
                id="01YCP46JKYM8FJCQ37NMBYHE6X",
                name="store2",
                created_at=datetime.fromisoformat("2022-07-25T21:15:37.524+00:00"),
                updated_at=datetime.fromisoformat("2022-07-25T21:15:37.524+00:00"),
                deleted_at=datetime.fromisoformat("2022-07-25T21:15:37.524+00:00"),
            )

            stores = [store1, store2]
            self.assertEqual(api_response.stores, stores)
            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores',
                headers=ANY,
                query_params=[('page_size', 1), ('continuation_token',
                                                 'continuation_token_example')],
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_create_store(self, mock_request):
        """Test case for create_store

        Create a store  # noqa: E501
        """
        response_body = '''{
            "id": "01YCP46JKYM8FJCQ37NMBYHE5X",
            "name": "test_store",
            "created_at": "2022-07-25T17:41:26.607Z",
            "updated_at": "2022-07-25T17:41:26.607Z"}
            '''
        mock_request.return_value = mock_response(response_body, 201)
        configuration = self.configuration
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.create_store(
                CreateStoreRequest(name="test-store"),
                options={}
            )
            self.assertIsInstance(api_response, CreateStoreResponse)
            self.assertEqual(api_response.id, '01YCP46JKYM8FJCQ37NMBYHE5X')
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"name": "test-store"},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_get_store(self, mock_request):
        """Test case for get_store

        Get all stores  # noqa: E501
        """
        response_body = '''
{
    "id": "01YCP46JKYM8FJCQ37NMBYHE5X",
    "name": "store1",
    "created_at": "2022-07-25T21:15:37.524Z",
    "updated_at": "2022-07-25T21:15:37.524Z"
}
            '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.get_store(
                options={}
            )
            self.assertIsInstance(api_response, GetStoreResponse)
            self.assertEqual(api_response.id, "01YCP46JKYM8FJCQ37NMBYHE5X")
            self.assertEqual(api_response.name, "store1")
            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_delete_store(self, mock_request):
        """Test case for delete_store

        Get all stores  # noqa: E501
        """
        mock_request.return_value = mock_response('', 201)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_client.delete_store(
                options={}
            )
            mock_request.assert_called_once_with(
                'DELETE',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X',
                headers=ANY,
                query_params=[],
                body=None,
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_read_authorization_models(self, mock_request):
        """Test case for read_authorization_models

        Return all authorization models configured for the store  # noqa: E501
        """
        response_body = '''
{
  "authorization_models": [{
    "id": "01G5JAVJ41T49E9TT3SKVS7X1J",
    "schema_version":"1.1",
    "type_definitions": [
      {
        "type": "document",
        "relations": {
          "reader": {
            "union": {
              "child": [
                {
                  "this": {}
                },
                {
                  "computedUserset": {
                    "object": "",
                    "relation": "writer"
                  }
                }
              ]
            }
          },
          "writer": {
            "this": {}
          }
        }
      }
    ]
  }],
  "continuation_token": "eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ=="
}
        '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        # Enter a context with an instance of the API client
        with OpenFgaClient(configuration) as api_client:

            # Return a particular version of an authorization model
            api_response = api_client.read_authorization_models(
                options={}
            )
            self.assertIsInstance(api_response, ReadAuthorizationModelsResponse)
            type_definitions = [
                TypeDefinition(
                    type="document",
                    relations=dict(
                        reader=Userset(
                            union=Usersets(
                                child=[
                                    Userset(this=dict()),
                                    Userset(computed_userset=ObjectRelation(
                                        object="",
                                        relation="writer",
                                    )),
                                ],
                            ),
                        ),
                        writer=Userset(
                            this=dict(),
                        ),
                    )
                )
            ]
            authorization_model = AuthorizationModel(id='01G5JAVJ41T49E9TT3SKVS7X1J', schema_version="1.1",
                                                     type_definitions=type_definitions)
            self.assertEqual(api_response.authorization_models, [authorization_model])
            self.assertEqual(api_response.continuation_token,
                             "eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ==")
            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write_authorization_model(self, mock_request):
        """Test case for write_authorization_model

        Create a new authorization model  # noqa: E501
        """
        response_body = '{"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}'
        mock_request.return_value = mock_response(response_body, 201)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            # example passing only required values which don't have defaults set
            body = WriteAuthorizationModelRequest(
                schema_version="1.1",
                type_definitions=[
                    TypeDefinition(
                        type="document",
                        relations=dict(
                            writer=Userset(
                                this=dict(),
                            ),
                            reader=Userset(
                                union=Usersets(
                                    child=[
                                        Userset(this=dict()),
                                        Userset(computed_userset=ObjectRelation(
                                            object="",
                                            relation="writer",
                                        )),
                                    ],
                                ),
                            ),
                        )
                    ),
                ],
            )
            # Create a new authorization model
            api_response = api_client.write_authorization_model(
                body,
                options={}
            )
            self.assertIsInstance(api_response, WriteAuthorizationModelResponse)
            expected_response = WriteAuthorizationModelResponse(
                authorization_model_id='01G5JAVJ41T49E9TT3SKVS7X1J'
            )
            self.assertEqual(api_response, expected_response)
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"schema_version": "1.1", "type_definitions": [{"type": "document", "relations": {"writer": {"this": {
                }}, "reader": {"union": {"child": [{"this": {}}, {"computedUserset": {"object": "", "relation": "writer"}}]}}}}]},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_read_authorization_model(self, mock_request):
        """Test case for read_authorization_model

        Return a particular version of an authorization model  # noqa: E501
        """
        response_body = '''
{
  "authorization_model": {
    "id": "01G5JAVJ41T49E9TT3SKVS7X1J",
    "schema_version":"1.1",
    "type_definitions": [
      {
        "type": "document",
        "relations": {
          "reader": {
            "union": {
              "child": [
                {
                  "this": {}
                },
                {
                  "computedUserset": {
                    "object": "",
                    "relation": "writer"
                  }
                }
              ]
            }
          },
          "writer": {
            "this": {}
          }
        }
      }
    ]
  }
}
        '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        # Enter a context with an instance of the API client
        with OpenFgaClient(configuration) as api_client:

            # Return a particular version of an authorization model
            api_response = api_client.read_authorization_model(
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}
            )
            self.assertIsInstance(api_response, ReadAuthorizationModelResponse)
            type_definitions = [
                TypeDefinition(
                    type="document",
                    relations=dict(
                        reader=Userset(
                            union=Usersets(
                                child=[
                                    Userset(this=dict()),
                                    Userset(computed_userset=ObjectRelation(
                                        object="",
                                        relation="writer",
                                    )),
                                ],
                            ),
                        ),
                        writer=Userset(
                            this=dict(),
                        ),
                    )
                )
            ]
            authorization_model = AuthorizationModel(id='01G5JAVJ41T49E9TT3SKVS7X1J', schema_version="1.1",
                                                     type_definitions=type_definitions)
            self.assertEqual(api_response.authorization_model, authorization_model)
            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_read_latest_authorization_model(self, mock_request):
        """Test case for read_latest_authorization_model

        Return the latest authorization models configured for the store  # noqa: E501
        """
        response_body = '''
{
  "authorization_models": [{
    "id": "01G5JAVJ41T49E9TT3SKVS7X1J",
    "schema_version":"1.1",
    "type_definitions": [
      {
        "type": "document",
        "relations": {
          "reader": {
            "union": {
              "child": [
                {
                  "this": {}
                },
                {
                  "computedUserset": {
                    "object": "",
                    "relation": "writer"
                  }
                }
              ]
            }
          },
          "writer": {
            "this": {}
          }
        }
      }
    ]
  }],
  "continuation_token": "eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ=="
}
        '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        # Enter a context with an instance of the API client
        with OpenFgaClient(configuration) as api_client:

            # Return a particular version of an authorization model
            api_response = api_client.read_latest_authorization_model(
                options={}
            )
            self.assertIsInstance(api_response, ReadAuthorizationModelResponse)
            type_definitions = [
                TypeDefinition(
                    type="document",
                    relations=dict(
                        reader=Userset(
                            union=Usersets(
                                child=[
                                    Userset(this=dict()),
                                    Userset(computed_userset=ObjectRelation(
                                        object="",
                                        relation="writer",
                                    )),
                                ],
                            ),
                        ),
                        writer=Userset(
                            this=dict(),
                        ),
                    )
                )
            ]
            authorization_model = AuthorizationModel(id='01G5JAVJ41T49E9TT3SKVS7X1J', schema_version="1.1",
                                                     type_definitions=type_definitions)
            self.assertEqual(api_response.authorization_model, authorization_model)
            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models',
                headers=ANY,
                query_params=[('page_size', 1)],
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_read_changes(self, mock_request):
        """Test case for read_changes

        Return a list of all the tuple changes  # noqa: E501
        """
        response_body = '''
{
  "changes": [
    {
      "tuple_key": {
        "object": "document:2021-budget",
        "relation": "reader",
        "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"
      },
      "operation": "TUPLE_OPERATION_WRITE",
      "timestamp": "2022-07-26T15:55:55.809Z"
    }
  ],
  "continuation_token": "eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ=="
}
        '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        # Enter a context with an instance of the API client
        with OpenFgaClient(configuration) as api_client:

            # Return a particular version of an authorization model
            api_response = api_client.read_changes(
                ClientReadChangesRequest("document"),
                options={"page_size": 1, "continuation_token": "abcdefg"}
            )

            self.assertIsInstance(api_response, ReadChangesResponse)
            changes = TupleChange(
                tuple_key=TupleKey(object="document:2021-budget", relation="reader",
                                   user="user:81684243-9356-4421-8fbf-a4f8d36aa31b"),
                operation=TupleOperation.WRITE,
                timestamp=datetime.fromisoformat("2022-07-26T15:55:55.809+00:00"))
            read_changes = ReadChangesResponse(
                continuation_token='eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ==',
                changes=[changes])
            self.assertEqual(api_response, read_changes)
            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/changes',
                headers=ANY,
                query_params=[('type', 'document'), ('page_size', 1),
                              ('continuation_token', 'abcdefg')],
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_read(self, mock_request):
        """Test case for read

        Get tuples from the store that matches a query, without following userset rewrite rules  # noqa: E501
        """
        response_body = '''
            {
  "tuples": [
    {
      "key": {
        "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b",
        "relation": "reader",
        "object": "document:2021-budget"
      },
      "timestamp": "2021-10-06T15:32:11.128Z"
    }
  ]
}
        '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        # Enter a context with an instance of the API client
        with OpenFgaClient(configuration) as api_client:
            body = TupleKey(
                object="document:2021-budget",
                relation="reader",
                user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
            )
            api_response = api_client.read(
                body=body,
                options={"page_size": 50, "continuation_token":
                         "eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ=="}
            )
            self.assertIsInstance(api_response, ReadResponse)
            key = TupleKey(user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                           relation="reader", object="document:2021-budget")
            timestamp = datetime.fromisoformat("2021-10-06T15:32:11.128+00:00")
            expected_data = ReadResponse(tuples=[Tuple(key=key, timestamp=timestamp)])
            self.assertEqual(api_response, expected_data)
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/read',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"},
                      "page_size": 50, "continuation_token": "eyJwayI6IkxBVEVTVF9OU0NPTkZJR19hdXRoMHN0b3JlIiwic2siOiIxem1qbXF3MWZLZExTcUoyN01MdTdqTjh0cWgifQ=="},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_read_empty_options(self, mock_request):
        """Test case for read with empty options

        Get tuples from the store that matches a query, without following userset rewrite rules  # noqa: E501
        """
        response_body = '''
            {
  "tuples": [
    {
      "key": {
        "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b",
        "relation": "reader",
        "object": "document:2021-budget"
      },
      "timestamp": "2021-10-06T15:32:11.128Z"
    }
  ]
}
        '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        # Enter a context with an instance of the API client
        with OpenFgaClient(configuration) as api_client:
            body = TupleKey(
                object="document:2021-budget",
                relation="reader",
                user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
            )
            api_response = api_client.read(
                body=body,
                options={}
            )
            self.assertIsInstance(api_response, ReadResponse)
            key = TupleKey(user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                           relation="reader", object="document:2021-budget")
            timestamp = datetime.fromisoformat("2021-10-06T15:32:11.128+00:00")
            expected_data = ReadResponse(tuples=[Tuple(key=key, timestamp=timestamp)])
            self.assertEqual(api_response, expected_data)
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/read',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget", "relation": "reader",
                                    "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write(self, mock_request):
        """Test case for write

        Add tuples from the store with transaction enabled
        """
        response_body = '{}'
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            body = ClientWriteRequest(
                writes=[
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                    )
                ],
            )
            api_client.write(
                body,
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}
            )
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, {"object": "document:2021-budget", "relation": "reader",
                                                                                                                                                                "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}, {"object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_delete(self, mock_request):
        """Test case for delete

        Delete tuples from the store with transaction enabled
        """
        response_body = '{}'
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            body = ClientWriteRequest(
                deletes=[
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    )
                ],
            )
            api_client.write(
                body,
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}
            )
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"deletes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                  "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write_batch(self, mock_request):
        """Test case for write

        Add tuples from the store with transaction disabled
        """
        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response('{}', 200),
            mock_response('{}', 200),
            mock_response('{}', 200),
        ]
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            body = ClientWriteRequest(
                writes=[
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                    )
                ],
            )
            transaction = WriteTransactionOpts(
                disabled=True, max_per_chunk=1, max_parallel_requests=10)
            response = api_client.write(
                body,
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J",
                         "transaction": transaction}
            )

            self.assertEqual(response.deletes, None)
            self.assertEqual(response.writes,
                             [
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                                     ),
                                     success=True,
                                     error=None),
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                                     ),
                                     success=True,
                                     error=None),
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                                     ),
                                     success=True,
                                     error=None)
                             ]
                             )
            self.assertEqual(mock_request.call_count, 4)
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write_batch_min_parallel(self, mock_request):
        """Test case for write

        Add tuples from the store with transaction disabled and minimum parallel request
        """
        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response('{}', 200),
            mock_response('{}', 200),
            mock_response('{}', 200),
        ]
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            body = ClientWriteRequest(
                writes=[
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                    )
                ],
            )
            transaction = WriteTransactionOpts(
                disabled=True, max_per_chunk=1, max_parallel_requests=1)
            response = api_client.write(
                body,
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J",
                         "transaction": transaction}
            )

            self.assertEqual(response.deletes, None)
            self.assertEqual(response.writes,
                             [
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                                     ),
                                     success=True,
                                     error=None),
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                                     ),
                                     success=True,
                                     error=None),
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                                     ),
                                     success=True,
                                     error=None)
                             ]
                             )
            self.assertEqual(mock_request.call_count, 4)
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write_batch_larger_chunk(self, mock_request):
        """Test case for write

        Add tuples from the store with transaction disabled and minimum parallel request
        """
        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response('{}', 200),
            mock_response('{}', 200),
        ]
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            body = ClientWriteRequest(
                writes=[
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                    )
                ],
            )
            transaction = WriteTransactionOpts(
                disabled=True, max_per_chunk=2, max_parallel_requests=2)
            response = api_client.write(
                body,
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J",
                         "transaction": transaction}
            )

            self.assertEqual(response.deletes, None)
            self.assertEqual(response.writes,
                             [
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                                     ),
                                     success=True,
                                     error=None),
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                                     ),
                                     success=True,
                                     error=None),
                                 ClientWriteSingleResponse(
                                     tuple_key=ClientTuple(
                                         object="document:2021-budget",
                                         relation="reader",
                                         user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                                     ),
                                     success=True,
                                     error=None)
                             ]
                             )
            self.assertEqual(mock_request.call_count, 3)
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, {
                    "object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write_batch_failed(self, mock_request):
        """Test case for write

        Add tuples from the store with transaction disabled where one of the request failed
        """
        response_body = '''
{
  "code": "validation_error",
  "message": "Generic validation error"
}
        '''

        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response('{}', 200),
            ValidationException(http_resp=http_mock_response(response_body, 400)),
            mock_response('{}', 200),
        ]
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            body = ClientWriteRequest(
                writes=[
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                    )
                ],
            )
            transaction = WriteTransactionOpts(
                disabled=True, max_per_chunk=1, max_parallel_requests=10)
            response = api_client.write(
                body,
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J",
                         "transaction": transaction}
            )

            self.assertEqual(response.deletes, None)
            self.assertEqual(len(response.writes), 3)
            self.assertEqual(response.writes[0],
                             ClientWriteSingleResponse(
                tuple_key=ClientTuple(
                    object="document:2021-budget",
                    relation="reader",
                    user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                ),
                success=True,
                error=None))
            self.assertEqual(response.writes[1].tuple_key,
                             ClientTuple(
                object="document:2021-budget",
                relation="reader",
                user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
            ))
            self.assertFalse(response.writes[1].success)
            self.assertIsInstance(response.writes[1].error, ValidationException)
            self.assertIsInstance(
                response.writes[1].error.parsed_exception, ValidationErrorMessageResponse)
            self.assertEqual(response.writes[2],
                             ClientWriteSingleResponse(
                tuple_key=ClientTuple(
                    object="document:2021-budget",
                    relation="reader",
                    user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                ),
                success=True,
                error=None))
            self.assertEqual(mock_request.call_count, 4)
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_delete_batch(self, mock_request):
        """Test case for delete

        Delete tuples from the store with transaction disabled but there is only 1 relationship tuple
        """
        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response('{}', 200),
        ]
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            body = ClientWriteRequest(
                deletes=[
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    )
                ],
            )
            transaction = WriteTransactionOpts(
                disabled=True, max_per_chunk=1, max_parallel_requests=10)
            api_client.write(
                body,
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J",
                         "transaction": transaction}
            )
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"deletes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader",
                                                  "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write_tuples(self, mock_request):
        """Test case for write tuples

        Add tuples from the store with transaction enabled
        """
        response_body = '{}'
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            api_client.write_tuples(
                [
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                    )
                ],
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}
            )
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"writes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, {"object": "document:2021-budget", "relation": "reader",
                                                                                                                                                                "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}, {"object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_delete_tuples(self, mock_request):
        """Test case for delete tuples

        Add tuples from the store with transaction enabled
        """
        response_body = '{}'
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:

            api_client.delete_tuples(
                [
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
                    ),
                    ClientTuple(
                        object="document:2021-budget",
                        relation="reader",
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
                    )
                ],
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}
            )
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/write',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"deletes": {"tuple_keys": [{"object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, {"object": "document:2021-budget", "relation": "reader",
                                                                                                                                                                 "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}, {"object": "document:2021-budget", "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}]}, "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"},
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write_batch_unauthorized(self, mock_request):
        """Test case for write with 401 response
        """

        mock_request.side_effect = UnauthorizedException(
            http_resp=http_mock_response('{}', 401)
        )
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            with self.assertRaises(UnauthorizedException) as api_exception:
                body = ClientWriteRequest(
                    writes=[
                        ClientTuple(
                            object="document:2021-budget",
                            relation="reader",
                            user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                        )
                    ],
                )
                transaction = WriteTransactionOpts(
                    disabled=True, max_per_chunk=1, max_parallel_requests=10)
                api_client.write(
                    body,
                    options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J",
                             "transaction": transaction}
                )

            self.assertIsInstance(api_exception.exception, UnauthorizedException)
            mock_request.assert_called()
            self.assertEqual(mock_request.call_count, 1)

            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_check(self, mock_request):
        """Test case for check

        Check whether a user is authorized to access an object  # noqa: E501
        """

        # First, mock the response
        response_body = '{"allowed": true, "resolution": "1234"}'
        mock_request.return_value = mock_response(response_body, 200)
        body = ClientCheckRequest(
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
            relation="reader",
            object="document:budget",
            contextual_tuples=[
                ClientTuple(
                    user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                    relation="writer",
                    object="document:budget",
                ),
            ],
        )
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.check(
                body=body,
                options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"}
            )
            self.assertIsInstance(api_response, CheckResponse)
            self.assertTrue(api_response.allowed)
            # Make sure the API was called with the right data
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b", "relation": "reader", "object": "document:budget"},
                      "contextual_tuples": {"tuple_keys": [{"user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b", "relation": "writer", "object": "document:budget"}]},
                      "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_check_config_auth_model(self, mock_request):
        """Test case for check

        Check whether a user is authorized to access an object and the auth model is already encoded in store  # noqa: E501
        """

        # First, mock the response
        response_body = '{"allowed": true, "resolution": "1234"}'
        mock_request.return_value = mock_response(response_body, 200)
        body = ClientCheckRequest(
            object="document:2021-budget",
            relation="reader",
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
        )
        configuration = self.configuration
        configuration.store_id = store_id
        configuration.authorization_model_id = "01GXSA8YR785C4FYS3C0RTG7B1"
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.check(
                body=body,
                options={}
            )
            self.assertIsInstance(api_response, CheckResponse)
            self.assertTrue(api_response.allowed)
            # Make sure the API was called with the right data
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_batch_check_single_request(self, mock_request):
        """Test case for check with single request

        Check whether a user is authorized to access an object  # noqa: E501
        """

        # First, mock the response
        response_body = '{"allowed": true, "resolution": "1234"}'
        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response(response_body, 200),
        ]
        body = ClientCheckRequest(
            object="document:2021-budget",
            relation="reader",
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
        )
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.batch_check(
                body=[body],
                options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"}
            )
            self.assertIsInstance(api_response, list)
            self.assertEqual(len(api_response), 1)
            self.assertEqual(api_response[0].error, None)
            self.assertTrue(api_response[0].allowed)
            self.assertEqual(api_response[0].request, body)
            # Make sure the API was called with the right data
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01GXSA8YR785C4FYS3C0RTG7B1',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_batch_check_multiple_request(self, mock_request):
        """Test case for check with multiple request

        Check whether a user is authorized to access an object  # noqa: E501
        """

        # First, mock the response
        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response('{"allowed": true, "resolution": "1234"}', 200),
            mock_response('{"allowed": false, "resolution": "1234"}', 200),
            mock_response('{"allowed": true, "resolution": "1234"}', 200),
        ]
        body1 = ClientCheckRequest(
            object="document:2021-budget",
            relation="reader",
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
        )
        body2 = ClientCheckRequest(
            object="document:2021-budget",
            relation="reader",
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
        )
        body3 = ClientCheckRequest(
            object="document:2021-budget",
            relation="reader",
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
        )
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.batch_check(
                body=[body1, body2, body3],
                options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1",
                         "max_parallel_requests": 2}
            )
            self.assertIsInstance(api_response, list)
            self.assertEqual(len(api_response), 3)
            self.assertEqual(api_response[0].error, None)
            self.assertTrue(api_response[0].allowed)
            self.assertEqual(api_response[0].request, body1)
            self.assertEqual(api_response[1].error, None)
            self.assertFalse(api_response[1].allowed)
            self.assertEqual(api_response[1].request, body2)
            self.assertEqual(api_response[2].error, None)
            self.assertTrue(api_response[2].allowed)
            self.assertEqual(api_response[2].request, body3)
            # Make sure the API was called with the right data
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01GXSA8YR785C4FYS3C0RTG7B1',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_batch_check_multiple_request_fail(self, mock_request):
        """Test case for check with multiple request with one request failed

        Check whether a user is authorized to access an object  # noqa: E501
        """
        response_body = '''
{
  "code": "validation_error",
  "message": "Generic validation error"
}
        '''

        # First, mock the response
        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response('{"allowed": true, "resolution": "1234"}', 200),
            ValidationException(http_resp=http_mock_response(response_body, 400)),
            mock_response('{"allowed": false, "resolution": "1234"}', 200),
        ]
        body1 = ClientCheckRequest(
            object="document:2021-budget",
            relation="reader",
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
        )
        body2 = ClientCheckRequest(
            object="document:2021-budget",
            relation="reader",
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31c",
        )
        body3 = ClientCheckRequest(
            object="document:2021-budget",
            relation="reader",
            user="user:81684243-9356-4421-8fbf-a4f8d36aa31d",
        )
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.batch_check(
                body=[body1, body2, body3],
                options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1",
                         "max_parallel_requests": 2}
            )
            self.assertIsInstance(api_response, list)
            self.assertEqual(len(api_response), 3)
            self.assertEqual(api_response[0].error, None)
            self.assertTrue(api_response[0].allowed)
            self.assertEqual(api_response[0].request, body1)
            self.assertFalse(api_response[1].allowed)
            self.assertEqual(api_response[1].request, body2)
            self.assertIsInstance(api_response[1].error, ValidationException)
            self.assertIsInstance(
                api_response[1].error.parsed_exception, ValidationErrorMessageResponse)
            self.assertEqual(api_response[2].error, None)
            self.assertFalse(api_response[2].allowed)
            self.assertEqual(api_response[2].request, body3)
            # Make sure the API was called with the right data
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01GXSA8YR785C4FYS3C0RTG7B1',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31c"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31d"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_expand(self, mock_request):
        """Test case for expand

        Expand all relationships in userset tree format, and following userset rewrite rules.  Useful to reason about and debug a certain relationship  # noqa: E501
        """
        response_body = '''{
            "tree": {"root": {"name": "document:budget#reader", "leaf": {"users": {"users": ["user:81684243-9356-4421-8fbf-a4f8d36aa31b"]}}}}}
            '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            body = ClientExpandRequest(
                object="document:budget",
                relation="reader",
            )
            api_response = api_client.expand(
                body=body,
                options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"}
            )
            self.assertIsInstance(api_response, ExpandResponse)
            curUsers = Users(users=["user:81684243-9356-4421-8fbf-a4f8d36aa31b"])
            leaf = Leaf(users=curUsers)
            node = Node(name="document:budget#reader", leaf=leaf)
            userTree = UsersetTree(node)
            expected_response = ExpandResponse(userTree)
            self.assertEqual(api_response, expected_response)
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/expand',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:budget", "relation": "reader"},
                      "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_list_objects(self, mock_request):
        """Test case for list_objects

        List objects  # noqa: E501
        """
        response_body = '''
{
  "objects": [
    "document:abcd1234"
  ]
}
            '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            body = ClientListObjectsRequest(
                type="document",
                relation="reader",
                user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
            )
            # Get all stores
            api_response = api_client.list_objects(
                body, options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"})
            self.assertIsInstance(api_response, ListObjectsResponse)
            self.assertEqual(api_response.objects, ['document:abcd1234'])
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/list-objects',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={'authorization_model_id': '01GXSA8YR785C4FYS3C0RTG7B1',
                      'type': 'document', 'relation': 'reader', 'user': 'user:81684243-9356-4421-8fbf-a4f8d36aa31b'},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_list_objects_contextual_tuples(self, mock_request):
        """Test case for list_objects

        List objects  # noqa: E501
        """
        response_body = '''
{
  "objects": [
    "document:abcd1234"
  ]
}
            '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            body = ClientListObjectsRequest(
                type="document",
                relation="reader",
                user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                contextual_tuples=[
                    ClientTuple(
                        user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                        relation="writer",
                        object="document:budget",
                    ),
                ],
            )
            # Get all stores
            api_response = api_client.list_objects(
                body, options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"})
            self.assertIsInstance(api_response, ListObjectsResponse)
            self.assertEqual(api_response.objects, ['document:abcd1234'])
            mock_request.assert_called_once_with(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/list-objects',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={'authorization_model_id': '01GXSA8YR785C4FYS3C0RTG7B1',
                      'type': 'document', 'relation': 'reader', 'user': 'user:81684243-9356-4421-8fbf-a4f8d36aa31b',
                      'contextual_tuples': {'tuple_keys': [{'object': 'document:budget', 'relation': 'writer', 'user': 'user:81684243-9356-4421-8fbf-a4f8d36aa31b'}]}},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_list_relations(self, mock_request):
        """Test case for list relations

        Check whether a user is authorized to access an object  # noqa: E501
        """

        # First, mock the response
        mock_request.side_effect = [
            mock_response('{}', 200),
            mock_response('{"allowed": true, "resolution": "1234"}', 200),
            mock_response('{"allowed": false, "resolution": "1234"}', 200),
            mock_response('{"allowed": true, "resolution": "1234"}', 200),
        ]
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.list_relations(
                body=ClientListRelationsRequest(user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                                                relations=["reader", "owner", "viewer"],
                                                object="document:2021-budget"),
                options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"}
            )
            self.assertEqual(api_response, ["reader", "viewer"])

            # Make sure the API was called with the right data
            mock_request.assert_any_call(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01GXSA8YR785C4FYS3C0RTG7B1',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "reader", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "owner", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            mock_request.assert_any_call(
                'POST',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/check',
                headers=ANY,
                query_params=[],
                post_params=[],
                body={"tuple_key": {"object": "document:2021-budget",
                                    "relation": "viewer", "user": "user:81684243-9356-4421-8fbf-a4f8d36aa31b"}, "authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"},
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_list_relations_unauthorized(self, mock_request):
        """Test case for list relations with 401 response
        """

        mock_request.side_effect = UnauthorizedException(
            http_resp=http_mock_response('{}', 401)
        )
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            with self.assertRaises(UnauthorizedException) as api_exception:
                api_client.list_relations(
                    body=ClientListRelationsRequest(user="user:81684243-9356-4421-8fbf-a4f8d36aa31b",
                                                    relations=["reader", "owner", "viewer"],
                                                    object="document:2021-budget"),
                    options={"authorization_model_id": "01GXSA8YR785C4FYS3C0RTG7B1"}
                )

            self.assertIsInstance(api_exception.exception, UnauthorizedException)
            mock_request.assert_called()
            self.assertEqual(mock_request.call_count, 1)

            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/authorization-models/01GXSA8YR785C4FYS3C0RTG7B1',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )
            api_client.close()

    @patch.object(rest.RESTClientObject, 'request')
    def test_read_assertions(self, mock_request):
        """Test case for read assertions

        """
        response_body = '''
{
  "authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J",
  "assertions": [
    {
      "tuple_key": {
        "object": "document:2021-budget",
        "relation": "reader",
        "user": "user:anne"
      },
      "expectation": true
    }
  ]
}
        '''
        mock_request.return_value = mock_response(response_body, 200)
        configuration = self.configuration
        configuration.store_id = store_id
        # Enter a context with an instance of the API client
        with OpenFgaClient(configuration) as api_client:
            api_response = api_client.read_assertions(
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}
            )
            self.assertEqual(api_response, ReadAssertionsResponse(
                authorization_model_id="01G5JAVJ41T49E9TT3SKVS7X1J",
                assertions=[Assertion(
                    tuple_key=TupleKey(object="document:2021-budget", relation="reader",
                                       user="user:anne"),
                    expectation=True,
                )]
            ))
            mock_request.assert_called_once_with(
                'GET',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/assertions/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                query_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_write_assertions(self, mock_request):
        """Test case for write assertions

        Get all stores  # noqa: E501
        """
        mock_request.return_value = mock_response('', 204)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_client.write_assertions(
                [ClientAssertion(user="user:anne", relation="reader",
                                 object="document:2021-budget", expectation=True)],
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}
            )
            mock_request.assert_called_once_with(
                'PUT',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/assertions/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                body={"assertions": [{"tuple_key": {"object": "document:2021-budget",
                                                    "relation": "reader", "user": "user:anne"}, "expectation": True}]},
                query_params=[],
                post_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_set_store_id(self, mock_request):
        """Test case for write assertions

        Get all stores  # noqa: E501
        """
        mock_request.return_value = mock_response('', 204)
        configuration = self.configuration
        configuration.store_id = store_id
        with OpenFgaClient(configuration) as api_client:
            api_client.set_store_id("01YCP46JKYM8FJCQ37NMBYHE5Y")

            api_client.write_assertions(
                [ClientAssertion(user="user:anne", relation="reader",
                                 object="document:2021-budget", expectation=True)],
                options={"authorization_model_id": "01G5JAVJ41T49E9TT3SKVS7X1J"}
            )
            self.assertEqual(api_client.get_store_id(), "01YCP46JKYM8FJCQ37NMBYHE5Y")
            mock_request.assert_called_once_with(
                'PUT',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5Y/assertions/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                body={"assertions": [{"tuple_key": {"object": "document:2021-budget",
                                                    "relation": "reader", "user": "user:anne"}, "expectation": True}]},
                query_params=[],
                post_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_config_auth_model(self, mock_request):
        """Test case for write assertions

        Get all stores  # noqa: E501
        """
        mock_request.return_value = mock_response('', 204)
        configuration = self.configuration
        configuration.store_id = store_id
        configuration.authorization_model_id = "01G5JAVJ41T49E9TT3SKVS7X1J"
        with OpenFgaClient(configuration) as api_client:
            api_client.write_assertions(
                [ClientAssertion(user="user:anne", relation="reader",
                                 object="document:2021-budget", expectation=True)],
                options={}
            )
            self.assertEqual(api_client.get_authorization_model_id(), "01G5JAVJ41T49E9TT3SKVS7X1J")
            mock_request.assert_called_once_with(
                'PUT',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/assertions/01G5JAVJ41T49E9TT3SKVS7X1J',
                headers=ANY,
                body={"assertions": [{"tuple_key": {"object": "document:2021-budget",
                                                    "relation": "reader", "user": "user:anne"}, "expectation": True}]},
                query_params=[],
                post_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )

    @patch.object(rest.RESTClientObject, 'request')
    def test_update_auth_model(self, mock_request):
        """Test case for write assertions

        Get all stores  # noqa: E501
        """
        mock_request.return_value = mock_response('', 204)
        configuration = self.configuration
        configuration.store_id = store_id
        configuration.authorization_model_id = "01G5JAVJ41T49E9TT3SKVS7X1J"
        with OpenFgaClient(configuration) as api_client:
            api_client.set_authorization_model_id("01G5JAVJ41T49E9TT3SKVS7X2J")

            api_client.write_assertions(
                [ClientAssertion(user="user:anne", relation="reader",
                                 object="document:2021-budget", expectation=True)],
                options={}
            )
            mock_request.assert_called_once_with(
                'PUT',
                'http://api.fga.example/stores/01YCP46JKYM8FJCQ37NMBYHE5X/assertions/01G5JAVJ41T49E9TT3SKVS7X2J',
                headers=ANY,
                body={"assertions": [{"tuple_key": {"object": "document:2021-budget",
                                                    "relation": "reader", "user": "user:anne"}, "expectation": True}]},
                query_params=[],
                post_params=[],
                _preload_content=ANY,
                _request_timeout=None
            )

    def test_configuration_store_id_invalid(self):
        """
        Test whether ApiValueError is raised if host has query
        """
        configuration = ClientConfiguration(
            api_host='localhost',
            api_scheme='http',
            store_id="abcd"
        )
        self.assertRaises(FgaValidationException, configuration.is_valid)

    def test_configuration_authorization_model_id_invalid(self):
        """
        Test whether ApiValueError is raised if host has query
        """
        configuration = ClientConfiguration(
            api_host='localhost',
            api_scheme='http',
            store_id="01H15K9J85050XTEDPVM8DJM78",
            authorization_model_id="abcd"
        )
        self.assertRaises(FgaValidationException, configuration.is_valid)



# coding: utf-8

# flake8: noqa

"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

__version__ = "0.3.0"

from openfga_sdk.client.client import OpenFgaClient
from openfga_sdk.client.configuration import ClientConfiguration

# import apis into sdk package
from openfga_sdk.api.open_fga_api import OpenFgaApi

# import ApiClient
from openfga_sdk.api_client import ApiClient
from openfga_sdk.configuration import Configuration
from openfga_sdk.exceptions import OpenApiException
from openfga_sdk.exceptions import FgaValidationException
from openfga_sdk.exceptions import ApiValueError
from openfga_sdk.exceptions import ApiKeyError
from openfga_sdk.exceptions import ApiAttributeError
from openfga_sdk.exceptions import ApiException
# import models into sdk package
from openfga_sdk.models.any import Any
from openfga_sdk.models.assertion import Assertion
from openfga_sdk.models.authorization_model import AuthorizationModel
from openfga_sdk.models.check_request import CheckRequest
from openfga_sdk.models.check_response import CheckResponse
from openfga_sdk.models.computed import Computed
from openfga_sdk.models.contextual_tuple_keys import ContextualTupleKeys
from openfga_sdk.models.create_store_request import CreateStoreRequest
from openfga_sdk.models.create_store_response import CreateStoreResponse
from openfga_sdk.models.difference import Difference
from openfga_sdk.models.error_code import ErrorCode
from openfga_sdk.models.expand_request import ExpandRequest
from openfga_sdk.models.expand_response import ExpandResponse
from openfga_sdk.models.get_store_response import GetStoreResponse
from openfga_sdk.models.internal_error_code import InternalErrorCode
from openfga_sdk.models.internal_error_message_response import InternalErrorMessageResponse
from openfga_sdk.models.leaf import Leaf
from openfga_sdk.models.list_objects_request import ListObjectsRequest
from openfga_sdk.models.list_objects_response import ListObjectsResponse
from openfga_sdk.models.list_stores_response import ListStoresResponse
from openfga_sdk.models.metadata import Metadata
from openfga_sdk.models.node import Node
from openfga_sdk.models.nodes import Nodes
from openfga_sdk.models.not_found_error_code import NotFoundErrorCode
from openfga_sdk.models.object_relation import ObjectRelation
from openfga_sdk.models.path_unknown_error_message_response import PathUnknownErrorMessageResponse
from openfga_sdk.models.read_assertions_response import ReadAssertionsResponse
from openfga_sdk.models.read_authorization_model_response import ReadAuthorizationModelResponse
from openfga_sdk.models.read_authorization_models_response import ReadAuthorizationModelsResponse
from openfga_sdk.models.read_changes_response import ReadChangesResponse
from openfga_sdk.models.read_request import ReadRequest
from openfga_sdk.models.read_response import ReadResponse
from openfga_sdk.models.relation_metadata import RelationMetadata
from openfga_sdk.models.relation_reference import RelationReference
from openfga_sdk.models.status import Status
from openfga_sdk.models.store import Store
from openfga_sdk.models.tuple import Tuple
from openfga_sdk.models.tuple_change import TupleChange
from openfga_sdk.models.tuple_key import TupleKey
from openfga_sdk.models.tuple_keys import TupleKeys
from openfga_sdk.models.tuple_operation import TupleOperation
from openfga_sdk.models.tuple_to_userset import TupleToUserset
from openfga_sdk.models.type_definition import TypeDefinition
from openfga_sdk.models.users import Users
from openfga_sdk.models.userset import Userset
from openfga_sdk.models.userset_tree import UsersetTree
from openfga_sdk.models.userset_tree_difference import UsersetTreeDifference
from openfga_sdk.models.userset_tree_tuple_to_userset import UsersetTreeTupleToUserset
from openfga_sdk.models.usersets import Usersets
from openfga_sdk.models.validation_error_message_response import ValidationErrorMessageResponse
from openfga_sdk.models.write_assertions_request import WriteAssertionsRequest
from openfga_sdk.models.write_authorization_model_request import WriteAuthorizationModelRequest
from openfga_sdk.models.write_authorization_model_response import WriteAuthorizationModelResponse
from openfga_sdk.models.write_request import WriteRequest



# coding: utf-8
"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""


class WriteTransactionOpts:
    """
    OpenFGA client write transaction info
    """

    def __init__(self, disabled: bool = False, max_per_chunk: int = 1, max_parallel_requests: int = 10):
        self._disabled = disabled
        self._max_per_chunk = max_per_chunk
        self._max_parallel_requests = max_parallel_requests

    @property
    def disabled(self):
        """
        Return disabled
        """
        return self._disabled

    @property
    def max_per_chunk(self):
        """
        Return max per chunk
        """
        return self._max_per_chunk

    @property
    def max_parallel_requests(self):
        """
        Return max parallel requests
        """
        return self._max_parallel_requests

    @disabled.setter
    def disabled(self, value):
        """
        Set disabled
        """
        self._disabled = value

    @max_per_chunk.setter
    def max_per_chunk(self, value):
        """
        Set max_per_chunk
        """
        self._max_per_chunk = value

    @max_parallel_requests.setter
    def max_parallel_requests(self, value):
        """
        Set max_parallel_requests
        """
        self._max_parallel_requests = value



# coding: utf-8
"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""


class ClientReadChangesRequest():
    """
    ClientReadChangesRequest encapsulates the parameters required to read changes
    """

    def __init__(self, type: str):
        self._type = type

    @property
    def type(self):
        """
        Return type
        """
        return self._type



# coding: utf-8
"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

from openfga_sdk.configuration import Configuration
from openfga_sdk.exceptions import FgaValidationException
from openfga_sdk.validation import is_well_formed_ulid_string


class ClientConfiguration(Configuration):
    """
    OpenFGA client configuration
    """

    def __init__(
            self,
            api_scheme="https",
            api_host=None,
            store_id=None,
            credentials=None,
            retry_params=None,
            authorization_model_id=None, ):
        super().__init__(api_scheme, api_host, store_id, credentials, retry_params)
        self._authorization_model_id = authorization_model_id

    def is_valid(self):
        super().is_valid()

        if self.authorization_model_id is not None and self.authorization_model_id != "" and is_well_formed_ulid_string(self.authorization_model_id) is False:
            raise FgaValidationException(
                "authorization_model_id ('%s') is not in a valid ulid format" % self.authorization_model_id)

    @property
    def authorization_model_id(self):
        return self._authorization_model_id

    @authorization_model_id.setter
    def authorization_model_id(self, value):
        self._authorization_model_id = value



# flake8: noqa

"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

from openfga_sdk.client.models.assertion import ClientAssertion
from openfga_sdk.client.models.batch_check_response import BatchCheckResponse
from openfga_sdk.client.models.check_request import ClientCheckRequest
from openfga_sdk.client.models.expand_request import ClientExpandRequest
from openfga_sdk.client.models.list_objects_request import ClientListObjectsRequest
from openfga_sdk.client.models.list_relations_request import ClientListRelationsRequest
from openfga_sdk.client.models.read_changes_request import ClientReadChangesRequest
from openfga_sdk.client.models.tuple import ClientTuple
from openfga_sdk.client.models.write_request import ClientWriteRequest
from openfga_sdk.client.models.write_response import ClientWriteResponse
from openfga_sdk.client.models.write_transaction_opts import WriteTransactionOpts



# coding: utf-8
"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

from openfga_sdk.client.models.check_request import ClientCheckRequest
from openfga_sdk.models.check_response import CheckResponse


class BatchCheckResponse():
    """
    BatchCheckResponse encapsulates the response for a single batch check
    """

    def __init__(self, allowed: bool, request: ClientCheckRequest, response: CheckResponse, error: Exception = None):
        self._allowed = allowed
        self._request = request
        self._response = response
        self._error = error

    @property
    def allowed(self):
        """
        Return whether request is allowed
        """
        return self._allowed

    @property
    def request(self):
        """
        Return original request
        """
        return self._request

    @property
    def response(self):
        """
        Return original request
        """
        return self._response

    @property
    def error(self):
        """
        Return error associated with batch request (if any)
        """
        return self._error

    def __str__(self):
        """
        Return the class string
        """
        return "allowed %s request %s error %s" % (self._allowed, self._request, self._error)



# coding: utf-8
"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

from openfga_sdk.client.models.tuple import ClientTuple

from typing import List


class ClientListRelationsRequest():
    """
    ClientListRelationsRequest encapsulates the parameters required for list all relations user have with object
    """

    def __init__(self, user: str, relations: List[str], object: str, contextual_tuples: List[ClientTuple] = None):
        self._user = user
        self._relations = relations
        self._object = object
        self._contextual_tuples = contextual_tuples

    @property
    def user(self):
        """
        Return user
        """
        return self._user

    @property
    def relations(self):
        """
        Return relations
        """
        return self._relations

    @property
    def object(self):
        """
        Return object
        """
        return self._object

    @property
    def contextual_tuples(self):
        """
        Return contextual_tuples
        """
        return self._contextual_tuples

    @user.setter
    def user(self, value):
        """
        Set user
        """
        self._user = value

    @relations.setter
    def relations(self, value):
        """
        Set relations
        """
        self._relations = value

    @object.setter
    def object(self, value):
        """
        Set object
        """
        self._object = value

    @contextual_tuples.setter
    def contextual_tuples(self, value):
        """
        Set contextual tuples
        """
        self._contextual_tuples = value



# coding: utf-8
"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

from openfga_sdk.sync.api_client import ApiClient
from openfga_sdk.sync.open_fga_api import OpenFgaApi
from openfga_sdk.client.configuration import ClientConfiguration
from openfga_sdk.client.models.assertion import ClientAssertion
from openfga_sdk.client.models.check_request import ClientCheckRequest, construct_check_request
from openfga_sdk.client.models.batch_check_response import BatchCheckResponse
from openfga_sdk.client.models.tuple import ClientTuple, convert_tuple_keys
from openfga_sdk.client.models.write_request import ClientWriteRequest
from openfga_sdk.client.models.write_response import ClientWriteResponse
from openfga_sdk.client.models.expand_request import ClientExpandRequest
from openfga_sdk.client.models.list_objects_request import ClientListObjectsRequest
from openfga_sdk.client.models.write_single_response import construct_write_single_response
from openfga_sdk.client.models.write_transaction_opts import WriteTransactionOpts
from openfga_sdk.client.models.read_changes_request import ClientReadChangesRequest
from openfga_sdk.exceptions import FgaValidationException
from openfga_sdk.models.assertion import Assertion
from openfga_sdk.models.check_request import CheckRequest
from openfga_sdk.models.contextual_tuple_keys import ContextualTupleKeys
from openfga_sdk.models.create_store_request import CreateStoreRequest
from openfga_sdk.models.expand_request import ExpandRequest
from openfga_sdk.models.list_objects_request import ListObjectsRequest
from openfga_sdk.models.read_authorization_model_response import ReadAuthorizationModelResponse
from openfga_sdk.models.read_request import ReadRequest
from openfga_sdk.models.tuple_key import TupleKey
from openfga_sdk.models.write_assertions_request import WriteAssertionsRequest
from openfga_sdk.models.write_authorization_model_request import WriteAuthorizationModelRequest
from openfga_sdk.models.write_request import WriteRequest
from openfga_sdk.validation import is_well_formed_ulid_string

import uuid
from typing import List
from concurrent.futures import ThreadPoolExecutor

CLIENT_METHOD_HEADER = "X-OpenFGA-Client-Method"
CLIENT_BULK_REQUEST_ID_HEADER = "X-OpenFGA-Client-Bulk-Request-Id"


def _chuck_array(array, max_size):
    """
    Helper function to chuck array into arrays of max_size
    """
    return [array[i * max_size:(i + 1) * max_size] for i in range((len(array) + max_size - 1) // max_size)]


def set_heading_if_not_set(options: dict[str, int | str], name: str, value: str):
    """
    Set heading to the value if it is not set
    """
    if options is None:
        options = {}
    headers = options.get("headers")
    if headers is None:
        headers = {}
    if headers.get(name) is None:
        headers[name] = value
    options["headers"] = headers
    return options


def options_to_kwargs(options: dict[str, int | str] = None):
    """
    Return kargs with continuation_token and page_size
    """
    kwargs = {}
    if options is not None:
        if options.get("page_size"):
            kwargs["page_size"] = options["page_size"]
        if options.get("continuation_token"):
            kwargs["continuation_token"] = options["continuation_token"]
        if options.get("headers"):
            kwargs["_headers"] = options["headers"]
        if options.get("retry_params"):
            kwargs["_retry_params"] = options["retry_params"]
    return kwargs


def options_to_transaction_info(options: dict[str, int | str] = None):
    """
    Return the transaction info
    """
    if options is not None and options.get("transaction"):
        return options["transaction"]
    return WriteTransactionOpts()


def _check_allowed(response: BatchCheckResponse):
    """
    Helper function to return whether the response is check is allowed
    """
    return response.allowed


class OpenFgaClient():
    """
    OpenFgaClient is the entry point for invoking calls against the OpenFGA API.
    """

    def __init__(self, configuration: ClientConfiguration):
        self._client_configuration = configuration
        self._api_client = ApiClient(configuration)
        self._api = OpenFgaApi(self._api_client)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self._api.close()

    def _get_authorization_model_id(self, options: object) -> str | None:
        """
        Return the authorization model ID if specified in the options.
        Otherwise return the authorization model ID stored in the client's configuration
        """
        authorization_model_id = self._client_configuration.authorization_model_id
        if options is not None and "authorization_model_id" in options:
            authorization_model_id = options["authorization_model_id"]
        if authorization_model_id is None or authorization_model_id == "":
            return None
        if is_well_formed_ulid_string(authorization_model_id) is False:
            raise FgaValidationException(
                "authorization_model_id ('%s') is not in a valid ulid format" % authorization_model_id)
        return authorization_model_id

    def set_store_id(self, value):
        """
        Update the store ID in the configuration
        """
        self._api_client.set_store_id(value)

    def get_store_id(self):
        """
        Return the store id (if any) store in the configuration
        """
        return self._api_client.get_store_id()

    def set_authorization_model_id(self, value):
        """
        Update the authorizaiton model id in the configuration
        """
        self._client_configuration.authorization_model_id = value

    def get_authorization_model_id(self):
        """
        Return the authorizaiton model id
        """
        return self._client_configuration.authorization_model_id

    def _check_valid_api_connection(self, options: dict[str, int | str]):
        """
        Checks that a connection with the given configuration can be established
        """
        authorization_model_id = self._get_authorization_model_id(options)
        if authorization_model_id is not None and authorization_model_id != "":
            self.read_authorization_model(options)
        else:
            self.read_latest_authorization_model(options)

    #################
    # Stores
    #################

    def list_stores(self, options: dict[str, int | str] = None):
        """
        List the stores in the system
        :param page_size(options) - Number of items returned per request
        :param continuation_token(options) - No continuation_token by default
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        # convert options to kargs
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "ListStores")
        kwargs = options_to_kwargs(options)
        api_response = self._api.list_stores(
            **kwargs,
        )
        return api_response

    def create_store(self, body: CreateStoreRequest, options: dict[str, int | str] = None):
        """
        Create the stores in the system
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "CreateStore")
        kwargs = options_to_kwargs(options)
        api_response = self._api.create_store(
            body,
            **kwargs
        )
        return api_response

    def get_store(self, options: dict[str, int | str] = None):
        """
        Get the store info in the system. Store id is from the configuration.
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "GetStore")
        kwargs = options_to_kwargs(options)
        api_response = self._api.get_store(
            **kwargs,
        )
        return api_response

    def delete_store(self, options: dict[str, int | str] = None):
        """
        Delete the store from the system. Store id is from the configuration.
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "DeleteStore")
        kwargs = options_to_kwargs(options)
        api_response = self._api.delete_store(
            **kwargs,
        )
        return api_response

    #######################
    # Authorization Models
    #######################

    def read_authorization_models(self, options: dict[str, int | str] = None):
        """
        Return all the authorization models for a particular store.
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "ReadAuthorizationModels")
        kwargs = options_to_kwargs(options)
        api_response = self._api.read_authorization_models(
            **kwargs,
        )
        return api_response

    def write_authorization_model(self, body: WriteAuthorizationModelRequest, options: dict[str, int | str] = None):
        """
        Write authorization model.
        :param body - WriteAuthorizationModelRequest
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "WriteAuthorizationModel")
        kwargs = options_to_kwargs(options)
        api_response = self._api.write_authorization_model(
            body,
            **kwargs,
        )
        return api_response

    def read_authorization_model(self, options: dict[str, int | str] = None):
        """
        Read an authorization model.
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "ReadAuthorizationModel")
        kwargs = options_to_kwargs(options)
        authorization_model_id = self._get_authorization_model_id(options)
        api_response = self._api.read_authorization_model(
            authorization_model_id,
            **kwargs,
        )
        return api_response

    def read_latest_authorization_model(self, options: dict[str, int | str] = None):
        """
        Convenient method of reading the latest authorizaiton model
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(
            options, CLIENT_METHOD_HEADER, "ReadLatestAuthoriationModel")
        options["page_size"] = 1
        api_response = self.read_authorization_models(options)
        return ReadAuthorizationModelResponse(api_response.authorization_models[0])

    #######################
    # Relationship Tuples
    #######################

    def read_changes(self, body: ClientReadChangesRequest, options: dict[str, str] = None):
        """
        Read changes for specified type
        :param body - the type we want to look for change
        :param page_size(options) - Number of items returned per request
        :param continuation_token(options) - No continuation_token by default
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "ReadChanges")
        kwargs = options_to_kwargs(options)
        kwargs["type"] = body.type
        api_response = self._api.read_changes(
            **kwargs,
        )
        return api_response

    def read(self, body: TupleKey, options: dict[str, str] = None):
        """
        Read changes for specified type
        :param body - the tuples we want to read
        :param page_size(options) - Number of items returned per request
        :param continuation_token(options) - No continuation_token by default
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "Read")
        page_size = None
        continuation_token = None
        if options:
            if options.get("page_size"):
                page_size = options.get("page_size")
                options.pop("page_size")
            if options.get("continuation_token"):
                continuation_token = options.get("continuation_token")
                options.pop("continuation_token")
        kwargs = options_to_kwargs(options)

        api_response = self._api.read(
            ReadRequest(
                tuple_key=body,
                page_size=page_size,
                continuation_token=continuation_token,
            ),
            **kwargs,
        )
        return api_response

    def _write_single_batch(self, batch: List[ClientTuple], is_write: bool, options: dict[str, str] = None):
        try:
            write_batch = None
            delete_batch = None
            if is_write:
                write_batch = batch
            else:
                delete_batch = batch
            self._write_with_transaction(ClientWriteRequest(
                writes=write_batch, deletes=delete_batch), options)
            return [construct_write_single_response(i, True, None) for i in batch]
        except Exception as err:
            return [construct_write_single_response(i, False, err) for i in batch]

    def _write_batches(self, tuple_keys: List[ClientTuple], transaction: WriteTransactionOpts, is_write: bool, options: dict[str, str] = None):
        """
        Internal function for write/delete batches
        """
        chunks = _chuck_array(tuple_keys, transaction.max_per_chunk)

        write_batches = _chuck_array(chunks, transaction.max_parallel_requests)
        batch_write_responses = []
        for write_batch in write_batches:
            response = [self._write_single_batch(i, is_write, options) for i in write_batch]
            flatten_list = [
                item for batch_single_response in response for item in batch_single_response]
            batch_write_responses.extend(flatten_list)

        return batch_write_responses

    def _write_with_transaction(self, body: ClientWriteRequest, options: dict[str, str] = None):
        """
        Write or deletes tuples
        """
        kwargs = options_to_kwargs(options)
        writes_tuple_keys = None
        deletes_tuple_keys = None
        if body.writes_tuple_keys:
            writes_tuple_keys = body.writes_tuple_keys
        if body.deletes_tuple_keys:
            deletes_tuple_keys = body.deletes_tuple_keys

        self._api.write(
            WriteRequest(
                writes=writes_tuple_keys,
                deletes=deletes_tuple_keys,
                authorization_model_id=self._get_authorization_model_id(options),
            ),
            **kwargs,
        )
        # any error will result in exception being thrown and not reached below code
        writes_response = None
        if body.writes:
            writes_response = [construct_write_single_response(i, True, None) for i in body.writes]
        deletes_response = None
        if body.deletes:
            deletes_response = [construct_write_single_response(
                i, True, None) for i in body.deletes]
        return ClientWriteResponse(writes=writes_response, deletes=deletes_response)

    def write(self, body: ClientWriteRequest, options: dict[str, str] = None):
        """
        Write or deletes tuples
        :param body - the write request
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "Writes")
        transaction = options_to_transaction_info(options)
        if not transaction.disabled:
            results = self._write_with_transaction(body, options)
            return results

        options = set_heading_if_not_set(options, CLIENT_BULK_REQUEST_ID_HEADER, str(uuid.uuid4()))
        # TODO: this should be run in parallel
        self._check_valid_api_connection(options)

        # otherwise, it is not a transaction and it is a batch write requests
        writes_response = None
        if body.writes:
            writes_response = self._write_batches(body.writes, transaction, True, options)
        deletes_response = None
        if body.deletes:
            deletes_response = self._write_batches(body.deletes, transaction, False, options)
        return ClientWriteResponse(writes=writes_response, deletes=deletes_response)

    def write_tuples(self, body: List[ClientTuple], options: dict[str, str] = None):
        """
        Convenient method for writing tuples
        :param body - the list of tuples we want to write
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "WriteTuples")
        result = self.write(ClientWriteRequest(body, None), options)
        return result

    def delete_tuples(self, body: List[ClientTuple], options: dict[str, str] = None):
        """
        Convenient method for deleteing tuples
        :param body - the list of tuples we want to delete
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "DeleteTuples")
        result = self.write(ClientWriteRequest(None, body), options)
        return result

    #######################
    # Relationship Queries
    #######################
    def check(self, body: ClientCheckRequest, options: dict[str, str] = None):  # noqa: E501
        """
        Check whether a user is authorized to access an object
        :param body - ClientCheckRequest defining check request
        :param authorization_model_id(options) - Overrides the authorization model id in the configuration
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "Check")

        kwargs = options_to_kwargs(options)

        req_body = CheckRequest(
            tuple_key=TupleKey(
                user=body.user,
                relation=body.relation,
                object=body.object,
            ),
            authorization_model_id=self._get_authorization_model_id(options),
        )
        if body.contextual_tuples:
            req_body.contextual_tuples = ContextualTupleKeys(
                tuple_keys=convert_tuple_keys(body.contextual_tuples)
            )
        api_response = self._api.check(
            body=req_body,
            **kwargs
        )
        return api_response

    def _single_batch_check(self, body: ClientCheckRequest, options: dict[str, str] = None):  # noqa: E501
        """
        Run a single batch request and return body in a SingleBatchCheckResponse
        :param body - ClientCheckRequest defining check request
        :param authorization_model_id(options) - Overrides the authorization model id in the configuration
        """
        try:
            api_response = self.check(body, options)
            return BatchCheckResponse(allowed=api_response.allowed, request=body, response=api_response, error=None)
        except Exception as err:
            return BatchCheckResponse(allowed=False, request=body, response=None, error=err)

    def batch_check(self, body: List[ClientCheckRequest], options: dict[str, str] = None):  # noqa: E501
        """
        Run a set of checks
        :param body - list of ClientCheckRequest defining check request
        :param authorization_model_id(options) - Overrides the authorization model id in the configuration
        :param max_parallel_requests(options) - Max number of requests to issue in parallel. Defaults to 10
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "BatchCheck")
        options = set_heading_if_not_set(options, CLIENT_BULK_REQUEST_ID_HEADER, str(uuid.uuid4()))

        # TODO: this should be run in parallel
        self._check_valid_api_connection(options)

        max_parallel_requests = 10
        if options is not None and "max_parallel_requests" in options:
            max_parallel_requests = options["max_parallel_requests"]

        batch_check_response = []

        def single_batch_check(request):
            return self._single_batch_check(request, options)

        with ThreadPoolExecutor(max_workers=max_parallel_requests) as executor:
            for response in executor.map(single_batch_check, body):
                batch_check_response.append(response)

        return batch_check_response

    def expand(self, body: ClientExpandRequest, options: dict[str, str] = None):  # noqa: E501
        """
        Run expand request
        :param body - list of ClientExpandRequest defining expand request
        :param authorization_model_id(options) - Overrides the authorization model id in the configuration
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "Expand")
        kwargs = options_to_kwargs(options)

        req_body = ExpandRequest(
            tuple_key=TupleKey(
                relation=body.relation,
                object=body.object,
            ),
            authorization_model_id=self._get_authorization_model_id(options),
        )
        api_response = self._api.expand(
            body=req_body,
            **kwargs
        )
        return api_response

    def list_objects(self, body: ClientListObjectsRequest, options: dict[str, str] = None):  # noqa: E501
        """
        Run list object request
        :param body - list object parameters
        :param authorization_model_id(options) - Overrides the authorization model id in the configuration
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "ListObjects")
        kwargs = options_to_kwargs(options)

        req_body = ListObjectsRequest(
            authorization_model_id=self._get_authorization_model_id(options),
            user=body.user,
            relation=body.relation,
            type=body.type,
        )
        if body.contextual_tuples:
            req_body.contextual_tuples = ContextualTupleKeys(
                tuple_keys=convert_tuple_keys(body.contextual_tuples)
            )
        api_response = self._api.list_objects(
            body=req_body,
            **kwargs
        )
        return api_response

    def list_relations(self, body: ClientListObjectsRequest, options: dict[str, str] = None):  # noqa: E501
        """
        Return all the relations for which user has a relationship with the object
        :param body - list relation request
        :param authorization_model_id(options) - Overrides the authorization model id in the configuration
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "ListRelations")
        options = set_heading_if_not_set(options, CLIENT_BULK_REQUEST_ID_HEADER, str(uuid.uuid4()))

        request_body = [construct_check_request(
            user=body.user, relation=i, object=body.object, contextual_tuples=body.contextual_tuples) for i in body.relations]
        result = self.batch_check(request_body, options)
        # need to filter with the allowed response
        result_iterator = filter(_check_allowed, result)
        result_list = list(result_iterator)
        return [i.request.relation for i in result_list]

    #######################
    # Assertions
    #######################

    def read_assertions(self, options: dict[str, str] = None):  # noqa: E501
        """
        Return the assertions
        :param authorization_model_id(options) - Overrides the authorization model id in the configuration
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "ReadAssertions")

        kwargs = options_to_kwargs(options)
        authorization_model_id = self._get_authorization_model_id(options)
        api_response = self._api.read_assertions(authorization_model_id, **kwargs)
        return api_response

    def write_assertions(self, body: List[ClientAssertion], options: dict[str, str] = None):  # noqa: E501
        """
        Upsert the assertions
        :param body - Write assertion request
        :param authorization_model_id(options) - Overrides the authorization model id in the configuration
        :param header(options) - Custom headers to send alongside the request
        :param retryParams(options) - Override the retry parameters for this request
        :param retryParams.maxRetry(options) - Override the max number of retries on each API request
        :param retryParams.minWaitInMs(options) - Override the minimum wait before a retry is initiated
        """
        options = set_heading_if_not_set(options, CLIENT_METHOD_HEADER, "WriteAssertions")
        kwargs = options_to_kwargs(options)
        authorization_model_id = self._get_authorization_model_id(options)

        def map_to_assertion(client_assertion: ClientAssertion):
            return Assertion(TupleKey(
                user=client_assertion.user,
                relation=client_assertion.relation,
                object=client_assertion.object,
            ), client_assertion.expectation)

        api_request_body = WriteAssertionsRequest(
            [map_to_assertion(client_assertion) for client_assertion in body])
        api_response = self._api.write_assertions(
            authorization_model_id, api_request_body, **kwargs)
        return api_response



"""
   Python SDK for OpenFGA

   API version: 0.1
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://discord.gg/8naAwJfWN6
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

from dataclasses import dataclass
from datetime import datetime, timedelta
import json
import typing
import urllib3
from urllib.parse import urlparse

from openfga_sdk.credentials import Credentials
from openfga_sdk.exceptions import AuthenticationError


class OAuth2Client:

    def __init__(
        self,
        credentials: Credentials
    ):
        self._credentials = credentials
        self._access_token = None
        self._access_expiry_time = None

    def _token_valid(self):
        """
        Return whether token is valid
        """
        if self._access_token is None or self._access_expiry_time is None:
            return False
        if self._access_expiry_time < datetime.now():
            return False
        return True

    def _obtain_token(self, client):
        """
        Perform OAuth2 and obtain token
        """
        configuration = self._credentials.configuration
        token_url = 'https://{}/oauth/token'.format(configuration.api_issuer)
        body = {
            'client_id': configuration.client_id,
            'client_secret': configuration.client_secret,
            'audience': configuration.api_audience,
            'grant_type': "client_credentials",
        }
        headers = urllib3.response.HTTPHeaderDict(
            {'Accept': 'application/json', 'Content-Type': 'application/json', 'User-Agent': 'openfga-sdk (python) 0.3.0'})
        raw_response = client.POST(token_url, headers=headers, body=body)
        if 200 <= raw_response.status <= 299:
            try:
                api_response = json.loads(raw_response.data)
            except:  # noqa: E722
                raise AuthenticationError(http_resp=raw_response)
            if not api_response.get('expires_in') or not api_response.get('access_token'):
                raise AuthenticationError(http_resp=raw_response)
            self._access_expiry_time = datetime.now() + timedelta(seconds=int(api_response.get('expires_in')))
            self._access_token = api_response.get('access_token')
        else:
            raise AuthenticationError(http_resp=raw_response)

    def get_authentication_header(self, client):
        """
        If configured, return the header for authentication
        """
        # check to see token is valid
        if not self._token_valid():
            # In this case, the token is not valid, we need to get the refresh the token
            self._obtain_token(client)
        return {'Authorization': 'Bearer {}'.format(self._access_token)}
